define(['exports'], function (exports) { 'use strict';

  function _defineProperty(obj, key, value) {
    if (key in obj) {
      Object.defineProperty(obj, key, {
        value: value,
        enumerable: true,
        configurable: true,
        writable: true
      });
    } else {
      obj[key] = value;
    }

    return obj;
  }

  class PropBuilder {
      constructor(input) {
          this.schemaName = input;
      }
      // eslint-disable-next-line class-methods-use-this
      build(schemas) {
          throw new Error("Method not implemented.");
      }
  }

  function mergeRecursive(a, b) {
    const result = Object.assign(Object.create(Object.getPrototypeOf(a)), a);

    for (const key in b) {
      if (!result[key]) result[key] = b[key];

      if (Array.isArray(result[key])) {
        if (!Array.isArray(b[key])) throw new Error('The two entities have mismatched structures and cannot be merged.');
        result[key] = b[key];
      } else if (typeof result[key] === 'object') {
        if (typeof b[key] !== 'object') throw new Error('The two entities have mismatched structures and cannot be merged.');
        result[key] = mergeRecursive(result[key], b[key]);
      } else {
        result[key] = b[key];
      }
    }

    return result;
  }

  /**
   * @class EntitySchema
   * @property {string} nameProp The name that will be used for normalized entities of this type
   * @property {Function} idFunction A function used to get the ID value for a single entity
   * @property {Function} processFunction A function used to convert the input type before normalizing
   * @property {Object<string, Schema>} props A map of properties of the entity type to the names of other schemas
   */

  class EntitySchema {
    /**
     * @type {string}
     */

    /**
     * @type {Function}
     */

    /**
     * @type {Function}
     */

    /**
     * @type {Object<string, string>}
     */
    constructor(name, idFunction, props, schemas, processFunction) {
      _defineProperty(this, "nameProp", void 0);

      _defineProperty(this, "idFunction", void 0);

      _defineProperty(this, "processFunction", void 0);

      _defineProperty(this, "props", void 0);

      this.nameProp = name;
      this.idFunction = idFunction;
      this.processFunction = processFunction;
      const allSchemas = { ...schemas,
        [name]: this
      }; // Convert the props into valid schemas

      const schemaProps = Object.entries(props).reduce((propsObject, [nextKey, nextValue]) => {
        if (typeof nextValue === 'string') {
          return { ...propsObject,
            [nextKey]: allSchemas[nextValue]
          };
        }

        return { ...propsObject,
          [nextKey]: nextValue.build(allSchemas)
        };
      }, {});
      this.props = schemaProps;
    }

    normalizeWith(input, parent, objectKey) {
      // Initialize the entities object to ensure that all entity types are defined, even if empty
      const baseEntities = Object.values(this.props).reduce((entities, nextSchema) => ({ ...entities,
        [nextSchema.nameProp]: {}
      }), {});
      const processedInput = this.processFunction(input, parent, objectKey);
      const {
        "entities": subEntities,
        processedObject
      } = Object.entries(processedInput).reduce(({
        entities,
        "processedObject": partialObject
      }, [inputKey, inputValue]) => {
        if (inputKey in this.props) {
          const schema = this.props[inputKey];
          const {
            "result": propResult,
            "entities": propEntities
          } = schema.normalizeWith(inputValue, processedInput, inputKey); // merge the current entities with the results of the normalize

          return {
            "entities": mergeRecursive(entities, propEntities),
            "processedObject": { ...partialObject,
              [inputKey]: propResult
            }
          };
        }

        return {
          entities,
          "processedObject": { ...partialObject,
            [inputKey]: inputValue
          }
        };
      }, {
        "entities": baseEntities,
        "processedObject": {}
      });
      const thisId = this.idFunction(processedInput, parent, objectKey);
      return {
        "result": thisId,
        "entities": mergeRecursive(subEntities, {
          [this.nameProp]: {
            [thisId]: processedObject
          }
        })
      };
    }

    normalize(input) {
      return this.normalizeWith(input, null);
    }

    normalizeManyWith(input, parent, objectKey) {
      const inputs = (() => {
        if (Array.isArray(input)) {
          return input.map(x => [objectKey, x]);
        } else if (typeof input === 'object') {
          return Object.entries(input);
        }

        throw new Error('Cannot process `normalizeMany` with a scalar input type');
      })();

      return inputs.reduce(({
        result,
        "entities": currentEntities
      }, [nextKey, nextVal]) => {
        const {
          "result": keyResult,
          "entities": keyEntities
        } = this.normalizeWith(nextVal, parent, nextKey);
        return {
          "result": result.concat([keyResult]),
          "entities": mergeRecursive(currentEntities, keyEntities)
        };
      }, {
        "result": [],
        "entities": {}
      });
    }

    normalizeMany(input) {
      return this.normalizeManyWith(input, null);
    }

  }
  /**
   * @class EntityBuilder
   * @property {Function} idFunction A function used to get the ID value for a single entity
   * @property {Function} processFunction A function used to convert the input type before normalizing
   * @property {string} nameProp The name that will be used for normalized entities of this type
   * @property {Object<string, string>} props A map of properties of the entity type to the names of other schemas
   * @property {Object<string, EntitySchema>} schemas A map of the names of other schemas to their schema classes
   */

  class EntityBuilder {
    /**
     * @type {string}
     */

    /**
     * @type {Function}
     */

    /**
     * @type {Function}
     */

    /**
     * @type {Object<string, string>}
     */

    /**
     * @type {Object<string, EntitySchema>}
     */
    constructor(id, name, props, schemas, processFunction) {
      _defineProperty(this, "nameProp", void 0);

      _defineProperty(this, "idFunction", void 0);

      _defineProperty(this, "processFunction", void 0);

      _defineProperty(this, "props", void 0);

      _defineProperty(this, "schemas", void 0);

      this.idFunction = id;
      this.nameProp = name;
      this.props = props;
      this.schemas = schemas;
      this.processFunction = processFunction;
    }

    name(name) {
      return new EntityBuilder(this.idFunction, name, this.props, this.schemas, this.processFunction);
    }

    id(id) {
      if (typeof id === 'function') {
        return new EntityBuilder(id, this.nameProp, this.props, this.schemas, this.processFunction);
      }

      return new EntityBuilder(input => `${input[id]}`, this.nameProp, this.props, this.schemas, this.processFunction);
    }

    prop(propName, propValue) {
      const [updatedProps, updatedSchemas] = (() => {
        if (propValue instanceof EntitySchema) {
          return [{ ...this.props,
            [propName]: propValue.nameProp
          }, { ...this.schemas,
            [propValue.nameProp]: propValue
          }];
        } else if (propValue instanceof PropBuilder) {
          if (propValue.schema) {
            return [// Remove the schema from the PropBuilder, and use just the name...
            { ...this.props,
              [propName]: propValue.basePropBuilder()
            }, // ...then put the schema back into the schemas object
            { ...this.schemas,
              [propValue.schemaName]: propValue.schema
            }];
          }

          return [{ ...this.props,
            [propName]: propValue
          }, this.schemas];
        }

        return [{ ...this.props,
          [propName]: propValue
        }, this.schemas];
      })();

      return new EntityBuilder(this.idFunction, this.nameProp, updatedProps, updatedSchemas, this.processFunction);
    }

    define(schema) {
      return new EntityBuilder(this.idFunction, this.nameProp, this.props, { ...this.schemas,
        [schema.nameProp]: schema
      }, this.processFunction);
    }

  }

  class ArrayValuesSchema {
    constructor(entitySchema) {
      _defineProperty(this, "nameProp", void 0);

      _defineProperty(this, "entitySchema", void 0);

      this.nameProp = entitySchema.nameProp;
      this.entitySchema = entitySchema;
    }

    normalizeWith(input, parent, objectKey) {
      if (!Array.isArray(input)) {
        throw new Error('Cannot process non-array input type as array values');
      }

      return input.reduce(({
        result,
        "entities": currentEntities
      }, nextVal) => {
        const {
          "result": keyResult,
          "entities": keyEntities
        } = this.entitySchema.normalizeWith(nextVal, parent, objectKey);
        return {
          "result": result.concat([keyResult]),
          "entities": mergeRecursive(currentEntities, keyEntities)
        };
      }, {
        "result": [],
        "entities": {}
      });
    }

  }
  class ArrayValuesBuilder extends PropBuilder {
    build(schemas) {
      return new ArrayValuesSchema(schemas[this.schemaName]);
    }

  }
  class ArrayValuesSchemaBuilder extends ArrayValuesBuilder {
    constructor(input) {
      super(input.nameProp);

      _defineProperty(this, "schema", void 0);

      this.schema = input;
    }

    basePropBuilder() {
      return new ArrayValuesBuilder(this.schemaName);
    }

  }

  function identity(input) {
    return input;
  }

  class ObjectValuesSchema {
    constructor(entitySchema) {
      _defineProperty(this, "nameProp", void 0);

      _defineProperty(this, "entitySchema", void 0);

      this.nameProp = entitySchema.nameProp;
      this.entitySchema = entitySchema;
    }

    normalizeWith(input, parent) {
      if (typeof input !== 'object') {
        throw new Error('Cannot process non-object input type as object values');
      }

      return Object.entries(input).reduce(({
        result,
        "entities": currentEntities
      }, [nextKey, nextVal]) => {
        const {
          "result": keyResult,
          "entities": keyEntities
        } = this.entitySchema.normalizeWith(nextVal, parent, nextKey);
        return {
          "result": { ...result,
            [nextKey]: keyResult
          },
          "entities": mergeRecursive(currentEntities, keyEntities)
        };
      }, {
        "result": {},
        "entities": {}
      });
    }

  }
  class ObjectValuesBuilder extends PropBuilder {
    build(schemas) {
      return new ObjectValuesSchema(schemas[this.schemaName]);
    }

  }
  class ObjectValuesSchemaBuilder extends ObjectValuesBuilder {
    constructor(input) {
      super(input.nameProp);

      _defineProperty(this, "schema", void 0);

      this.schema = input;
    }

    basePropBuilder() {
      return new ObjectValuesBuilder(this.schemaName);
    }

  }

  function entity(processFunction) {
    if (!processFunction) {
      return new EntityBuilder(null, null, {}, {}, identity);
    }

    return new EntityBuilder(null, null, {}, {}, processFunction);
  }
  function arrayValues(schemaType) {
    if (typeof schemaType === 'string') {
      return new ArrayValuesBuilder(schemaType);
    }

    return new ArrayValuesSchemaBuilder(schemaType);
  }
  function objectValues(schemaType) {
    if (typeof schemaType === 'string') {
      return new ObjectValuesBuilder(schemaType);
    }

    return new ObjectValuesSchemaBuilder(schemaType);
  }
  function buildSchema(builder) {
    return new EntitySchema(builder.nameProp, builder.idFunction, builder.props, builder.schemas, builder.processFunction);
  }

  exports.arrayValues = arrayValues;
  exports.buildSchema = buildSchema;
  exports.entity = entity;
  exports.objectValues = objectValues;

  Object.defineProperty(exports, '__esModule', { value: true });

});
